from starlette.request import Request
from starlette.response import Response, PlainTextResponse
from starlette.types import Receive, Send, Scope


class View:
    def __init__(self, scope: Scope):
        self.scope = scope

    async def __call__(self, receive: Receive, send: Send):
        request = Request(self.scope, receive=receive)
        kwargs = self.scope.get("kwargs", {})
        response = await self.dispatch(request, **kwargs)
        await response(receive, send)

    async def dispatch(self, request: Request, **kwargs) -> Response:
        handler_name = "get" if request.method == "HEAD" else request.method.lower()
        handler = getattr(self, handler_name, self.method_not_allowed)
        return await handler(request, **kwargs)

    async def method_not_allowed(self, request: Request, **kwargs) -> Response:
        return PlainTextResponse("Method not allowed", 406)



class Middleware:
    def __init__(self, app):
        self.app = app

    def __call__(self, scope):
        async def asgi(receive, send):
            await self.handler(scope, receive, send)

        return asgi

    async def handler(self, scope, receive, send):
        raise NotImplemented()  # pragma: nocover


class ExceptionMiddleware:
    def __init__(self, app, handled_exceptions=None):
        self.app = app
        self.handled_exceptions = handled_exceptions or {}

    def __call__(self, scope):
        async def app(receive, send):
            try:
                instance = self.app(scope)
                await instance(receive, send)
            except Exception as exc:
                scope["exc"] = exc
                for cls, handler in self.handled_exceptions.items():
                    if isinstance(exc, cls):
                        instance = handler(scope)
                        await instance(receive, send)
                        return
                raise exc from None

        return app

    async def handler(self, scope, receive, send):
        raise NotImplemented()  # pragma: nocover



from starlette.datastructures import Headers
from starlette.response import HTMLResponse, PlainTextResponse
import html
import traceback


class DebugMiddleware:
    def __init__(self, app):
        self.app = app

    def __call__(self, scope):
        if scope["type"] != "http":
            return self.app(scope)
        return _DebugResponder(self.app, scope)


class _DebugResponder:
    def __init__(self, app, scope):
        self.app = app
        self.scope = scope
        self.response_started = False

    async def __call__(self, receive, send):
        self.raw_send = send
        try:
            asgi = self.app(self.scope)
            await asgi(receive, self.send)
        except:
            if self.response_started:
                raise
            headers = Headers(self.scope.get("headers", []))
            accept = headers.get("accept", "")
            if "text/html" in accept:
                exc_html = html.escape(traceback.format_exc())
                content = f"<html><body><h1>500 Server Error</h1><pre>{exc_html}</pre></body></html>"
                response = HTMLResponse(content, status_code=500)
            else:
                content = traceback.format_exc()
                response = PlainTextResponse(content, status_code=500)
            await response(receive, send)

    async def send(self, message):
        if message["type"] == "http.response.start":
            self.response_started = True
        await self.raw_send(message)



from starlette.request import Request
from starlette.routing import Path, PathPrefix, Router
from starlette.types import ASGIApp, ASGIInstance, Receive, Scope, Send
from starlette.websockets import WebSocketSession
import asyncio
import inspect


def request_response(func):
    """
    Takes a function or coroutine `func(request, **kwargs) -> response`,
    and returns an ASGI application.
    """
    is_coroutine = asyncio.iscoroutinefunction(func)

    def app(scope: Scope) -> ASGIInstance:
        async def awaitable(receive: Receive, send: Send) -> None:
            request = Request(scope, receive=receive)
            kwargs = scope.get("kwargs", {})
            if is_coroutine:
                response = await func(request, **kwargs)
            else:
                response = func(request, **kwargs)
            await response(receive, send)

        return awaitable

    return app


def websocket_session(func):
    """
    Takes a coroutine `func(session, **kwargs)`, and returns an ASGI application.
    """

    def app(scope: Scope) -> ASGIInstance:
        async def awaitable(receive: Receive, send: Send) -> None:
            session = WebSocketSession(scope, receive=receive, send=send)
            kwargs = scope.get("kwargs", {})
            await func(session, **kwargs)

        return awaitable

    return app


class App:
    def __init__(self) -> None:
        self.router = Router(routes=[])

    def mount(self, path: str, app: ASGIApp):
        prefix = PathPrefix(path, app=app)
        self.router.routes.append(prefix)

    def add_route(self, path: str, route, methods=None) -> None:
        if not inspect.isclass(route):
            route = request_response(route)
            if methods is None:
                methods = ["GET"]

        instance = Path(path, route, protocol="http", methods=methods)
        self.router.routes.append(instance)

    def add_websocket_route(self, path: str, route) -> None:
        if not inspect.isclass(route):
            route = websocket_session(route)

        instance = Path(path, route, protocol="websocket")
        self.router.routes.append(instance)

    def route(self, path: str):
        def decorator(func):
            self.add_route(path, func)
            return func

        return decorator

    def websocket_route(self, path: str):
        def decorator(func):
            self.add_websocket_route(path, func)
            return func

        return decorator

    def __call__(self, scope: Scope) -> ASGIInstance:
        return self.router(scope)



from starlette.app import App
from starlette.response import (
    FileResponse,
    HTMLResponse,
    JSONResponse,
    RedirectResponse,
    Response,
    PlainTextResponse,
    StreamingResponse,
)
from starlette.request import Request
from starlette.testclient import TestClient


__all__ = (
    "App",
    "FileResponse",
    "HTMLResponse",
    "JSONResponse",
    "PlainTextResponse",
    "RedirectResponse",
    "Response",
    "StreamingResponse",
    "Request",
    "TestClient",
)
__version__ = "0.2.1"
